#ifndef _XMA_PLG_TST_H_
#define _XMA_PLG_TST_H_

#define XMA_PLG_INIT    1
#define XMA_PLG_CLOSE   2
#define XMA_PLG_SEND    4
#define XMA_PLG_RECV    8

#define XMA_PLG_ENC     16
#define XMA_PLG_DEC     32
#define XMA_PLG_SCAL    64
#define XMA_PLG_FIL     128
#define XMA_PLG_KERN    256

#endif
